import numpy as np
from numpy.typing import ArrayLike
from abc import ABC, abstractmethod
from data.dataLoader import DataLoader


def mapTo(values: np.ndarray, arange: list = [0,1]) -> np.ndarray:
    assert len(arange) == 2, 'arange must be of length 2'
    assert arange[0] < arange[1], 'arange must start at a lower value than it ends'

    c, d = arange[0], arange[1]
    a, b = np.min(values), np.max(values)

    return c + ((d - c) / (b - a)) * (values - a)


class Topology(ABC):
    def __init__(self, gridSize: tuple[int, int], numFeatures: int):
        self.name = self.__class__.__name__
        self.stepCorrection = 0

        # basic parameters of the topology
        self.gridSize = list(gridSize)
        self.gridIndices = [] # used for visualizing the map as a scatter plot
        self.numFeatures = numFeatures
        self.numNeurons = np.prod(gridSize)

        # neighborhood relations
        self.neighborIndices = []
        self.neuronIndices = []

        # weights and neuron init
        self.weights = np.zeros((self.numNeurons, self.numFeatures))
        self.genIndices()

        # these arrays will be used assessing the topology
        self._movementCounts = np.zeros((self.numNeurons), dtype=int)
        self._directionCounts = np.zeros((self.numNeurons, self.numFeatures))
        self._travelDistance = np.zeros((self.numNeurons, self.numFeatures))
        self._minimalDistances = np.zeros((self.numNeurons, self.numFeatures))

    def initWeights(self, data: DataLoader | np.ndarray) -> None:
        if isinstance(data, DataLoader):
            minValues = np.min(data.dataSet, axis=0)
            maxValues = np.max(data.dataSet, axis=0)
        else:
            minValues = np.min(data, axis=0)
            maxValues = np.max(data, axis=0)

        # Initialize weights within the range for each feature
        for i in range(self.numFeatures):
            self.weights[:, i] = np.random.uniform(minValues[i], maxValues[i], self.numNeurons)

    @property
    def weightMatrix(self) -> np.ndarray:
        return np.linalg.norm(self.weights,axis=-1).reshape(*self.gridSize)

    @property
    def uMatrix(self) -> np.ndarray:
        matrix = np.zeros(self.numNeurons)
        for index, weight in enumerate(self.weights):
            distances = np.linalg.norm(weight - self.weights[self.getNeighbors(index)], axis=-1)
            matrix[index] = np.mean(distances)

        return matrix.reshape(*self.gridSize)

    def __getitem__(self, index: int | ArrayLike) -> np.ndarray:
        if isinstance(index, int | np.integer):
            return self.weights[index]
        index = np.array(index)
        if len(index.shape) == 2:
            index = np.ravel_multi_index(index, self.gridSize)
        return self.weights[index]

    def __setitem__(self, index: int | ArrayLike, value) -> None:
        if isinstance(index, int | np.integer):
            self.weights[index] = np.array(value)
            return
        index = np.array(index)
        if len(index.shape) == 2:
            index = np.ravel_multi_index(index, self.gridSize)
        self.weights[index] = np.array(value)

    def __iter__(self):
        return iter(self.weights)

    @abstractmethod
    def genIndices(self) -> None:
        raise NotImplementedError('this has not been implemented')

    def getNeighbors(self, index) -> np.ndarray:
        # checking if index is a single number
        if isinstance(index, (int, np.integer)):
            index = np.array([index])

        # broadcasting indices arrays
        tiledIndices = np.tile(self.neuronIndices, len(index))
        repeatedIndex = np.repeat(index, len(self.neuronIndices))

        # finding the equal values
        bools = (tiledIndices == repeatedIndex).reshape(len(index), len(self.neuronIndices))
        return self.neighborIndices[np.any(bools, axis=0)]

    def countMovements(self, bestMatches, neighbors, directions, neighborDirection) -> None:
        np.add.at(self._movementCounts, bestMatches, 1)
        np.add.at(self._movementCounts, neighbors, 1)
        self._directionCounts[bestMatches] += np.sign(directions).astype('int')
        self._directionCounts[neighbors] += np.sign(neighborDirection).astype('int')
        #self._travelDistance[bestMatches] += directions
        #self._travelDistance[neighbors] += neighborDirection

    def grow(self):
        weightsShape = list((*self.gridSize, self.numFeatures))
        movementsShape = list(self.gridSize)

        movements = mapTo(self._movementCounts)
        directions = np.sum(abs(self._directionCounts),axis=1)
        directions = abs(mapTo(directions, [-1,0]))
        one, two = np.argpartition(movements + directions, -2)[-2:]
        x1, y1 = np.unravel_index(one, shape=(self.gridSize))
        x2, y2 = np.unravel_index(two, shape=(self.gridSize))

        xDiff = abs(x1 - x2)
        yDiff = abs(y1 - y2)

        if xDiff > yDiff == 0:
            weightsShape[1] += 1
            movementsShape[1] += 1
        else:
            weightsShape[0] += 1
            movementsShape[0] += 1

        newMovements = np.zeros(movementsShape, dtype=int)
        newWeights = np.zeros(weightsShape)
        newDirections = np.zeros(weightsShape, dtype=int)

        if xDiff > yDiff == 0:
            for i in range(newMovements.shape[0]):
                weightLine = self.weights.reshape(*self.gridSize, self.numFeatures)[i]
                movementLine = self._movementCounts.reshape(*self.gridSize)[i]
                directionLine = self._directionCounts.reshape(*self.gridSize, self.numFeatures)[i]

                mappedMovements = mapTo(movementLine)
                mappedDirections = abs(mapTo(np.sum(abs(directionLine)), [-1,0]))

                maxIndex = np.argmax(mappedMovements + mappedDirections)

                if 0 < maxIndex < len(movementLine) - 1:
                    left = movementLine[maxIndex - 1]
                    right = movementLine[maxIndex + 1]
                    if left > right:
                        maxIndex -= 1

                if maxIndex == 0:
                    maxIndex += 1

                newMovementLine = np.insert(movementLine, maxIndex, 0)
                value = (weightLine[maxIndex-1] + weightLine[maxIndex])/2
                newWeightLine = np.insert(weightLine, maxIndex, value).reshape(-1,self.numFeatures)
                newDirectionLine = np.insert(directionLine, maxIndex, np.zeros(self.numFeatures,dtype=int)).reshape(-1,self.numFeatures)

                newMovements[i] = newMovementLine
                newWeights[i] = newWeightLine
                newDirections[i] = newDirectionLine
        else:
            for i in range(newMovements.shape[1]):
                weightLine = self.weights.reshape(*self.gridSize, self.numFeatures)[:,i]
                movementLine = self._movementCounts.reshape(*self.gridSize)[:,i]
                directionLine = self._directionCounts.reshape(*self.gridSize, self.numFeatures)[:,i]

                mappedMovements = mapTo(movementLine)
                mappedDirections = abs(mapTo(np.sum(abs(directionLine)), [-1,0]))

                maxIndex = np.argmax(mappedMovements + mappedDirections)

                if 0 < maxIndex < len(movementLine) - 1:
                    left = movementLine[maxIndex - 1]
                    right = movementLine[maxIndex + 1]
                    if left > right:
                        maxIndex -= 1

                if maxIndex == 0:
                    maxIndex += 1

                newMovementLine = np.insert(movementLine, maxIndex, 0)
                value = (weightLine[maxIndex-1] + weightLine[maxIndex])/2
                newWeightLine = np.insert(weightLine, maxIndex, value).reshape(-1,self.numFeatures)
                newDirectionLine = np.insert(directionLine, maxIndex, np.zeros(self.numFeatures,dtype=int)).reshape(-1,self.numFeatures)

                newMovements[:,i] = newMovementLine
                newWeights[:,i] = newWeightLine
                newDirections[:,i] = newDirectionLine

        self.gridSize = movementsShape
        self.numNeurons = np.prod(movementsShape)
        self._movementCounts = newMovements.reshape(self.numNeurons)
        self.weights = newWeights.reshape(self.numNeurons, self.numFeatures)
        self._directionCounts = newDirections.reshape(self.numNeurons, self.numFeatures)
        self.genIndices()

class Rectangular(Topology):
    def __init__(self, gridSize: tuple[int, int], numFeatures: int):
        super().__init__(gridSize, numFeatures)

    def genIndices(self) -> None:
        self.gridIndices = []
        self.neighborIndices = []
        self.neuronIndices = []

        # Link the Neurons together in a grid pattern
        for index in range(self.numNeurons):
            position = [index // self.gridSize[0], index % self.gridSize[1]]
            self.gridIndices.append(position)

            neighbors = []
            if index >= self.gridSize[1]:
                neighbors.append(index-self.gridSize[1])  # up
            if index < self.numNeurons - self.gridSize[1]:
                neighbors.append(index+self.gridSize[1])  # down
            if index % self.gridSize[1] != 0:
                neighbors.append(index-1)  # left
            if (index+1) % self.gridSize[1] != 0:
                neighbors.append(index+1)  # right

            self.neighborIndices.extend(neighbors)
            self.neuronIndices.extend([index] * len(neighbors))

        self.gridIndices = np.array(self.gridIndices)
        self.neighborIndices = np.array(self.neighborIndices)
        self.neuronIndices = np.array(self.neuronIndices)


class Hexagonal(Topology):
    def __init__(self, gridSize: tuple[int, int], numFeatures: int):
        super().__init__(gridSize, numFeatures)
        self.stepCorrection = 0.5

    def genIndices(self) -> None:
        self.gridIndices = []
        self.neighborIndices = []
        self.neuronIndices = []

        # Link the Neurons together in a hexagonal pattern
        for index in range(self.numNeurons):
            position = [index // self.gridSize[0], index % self.gridSize[1]]
            if position[0] % 2 == 0:
               position[1] += 0.5
            self.gridIndices.append(position)

            neighbors = []
            if index >= self.gridSize[1]:
                neighbors.append(index-self.gridSize[1])  # up
            if index < self.numNeurons - self.gridSize[1]:
                neighbors.append(index+self.gridSize[1])  # down
            if index % self.gridSize[1] != 0:
                neighbors.append(index-1)  # left
            if (index+1) % self.gridSize[1] != 0:
                neighbors.append(index+1)  # right

            # diagonal neighbors (based on even or odd row)
            row = index // self.gridSize[1]
            if row % 2 == 0:  # even rows
                if index >= self.gridSize[1] and (index+1) % self.gridSize[1] != 0:
                    neighbors.append(index-self.gridSize[1]+1)  # upper right
                if index < self.numNeurons - self.gridSize[1] and (index+1) % self.gridSize[1] != 0:
                    neighbors.append(index+self.gridSize[1]+1)  # lower right
            else:  # odd rows
                if index >= self.gridSize[1] and index % self.gridSize[1] != 0:
                    neighbors.append(index-self.gridSize[1]-1)  # upper left
                if index < self.numNeurons - self.gridSize[1] and index % self.gridSize[1] != 0:
                    neighbors.append(index+self.gridSize[1]-1)  # lower left

            self.neighborIndices.extend(neighbors)
            self.neuronIndices.extend([index] * len(neighbors))

        self.gridIndices = np.array(self.gridIndices)
        self.neighborIndices = np.array(self.neighborIndices)